<!DOCTYPE html>
<html>
  	<head>
    	<title>Khoj: Processor Settings</title>
    	<link rel=”stylesheet” href=”static/styles.css”>
    	<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/@picocss/pico@1/css/pico.min.css">
  	</head>
  	<body class="data-integration">
    	<header class=”header”>
      	<h1>Configure your processor integrations for Khoj</h1>
    	</header>
		<a href="/config">Go back</a>

    	<div class=”content”>
      		{% block content %}
      		{% endblock %}
    	</div>

    	<footer class=”footer”>
    	</footer>
  	</body>
	<style>
		body.data-integration {
			padding: 0 10%
		}
	</style>
</html>
